from copy import deepcopy
from datetime import date, datetime
from typing import Optional, Union

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.formats import date_format, time_format
from django.utils.functional import classproperty
from django.utils.translation import gettext_lazy as _

from calendarweek import CalendarWeek
from calendarweek.django import i18n_day_abbr_choices_lazy, i18n_day_name_choices_lazy
from recurrence.fields import RecurrenceField

from aleksis.apps.chronos.models import LessonEvent, SupervisionEvent
from aleksis.apps.cursus.models import Course, Subject
from aleksis.core.managers import CurrentSiteManagerWithoutMigrations
from aleksis.core.mixins import ExtensibleModel, ExtensiblePolymorphicModel
from aleksis.core.models import Holiday, Person, Room, SchoolTerm

from .managers import ValidityRangeQuerySet


class ValidityRange(ExtensibleModel):
    """A validity range is a date range in which certain data are valid."""

    objects = CurrentSiteManagerWithoutMigrations.from_queryset(ValidityRangeQuerySet)()

    school_term = models.ForeignKey(
        SchoolTerm,
        on_delete=models.CASCADE,
        verbose_name=_("School term"),
        related_name="lr_validity_ranges",
    )
    name = models.CharField(verbose_name=_("Name"), max_length=255, blank=True)

    date_start = models.DateField(verbose_name=_("Start date"))
    date_end = models.DateField(verbose_name=_("End date"))

    @classmethod
    def get_current(cls, day: Optional[date] = None) -> Optional["ValidityRange"]:
        """Get the currently active validity range."""
        if not day:
            day = timezone.now().date()
        try:
            return cls.objects.on_day(day).first()
        except ValidityRange.DoesNotExist:
            return None

    @classproperty
    def current(cls) -> Optional["ValidityRange"]:
        """Get the currently active validity range."""
        return cls.get_current()

    def clean(self):
        """Ensure that there is only one validity range at each point of time."""
        if self.date_end < self.date_start:
            raise ValidationError(_("The start date must be earlier than the end date."))

        if self.school_term:
            if (
                self.date_end > self.school_term.date_end
                or self.date_start < self.school_term.date_start
            ):
                raise ValidationError(_("The validity range must be within the school term."))

        qs = ValidityRange.objects.within_dates(self.date_start, self.date_end)
        if self.pk:
            qs = qs.exclude(pk=self.pk)
        if qs.exists():
            raise ValidationError(
                _("There is already a validity range for this time or a part of this time.")
            )

    def __str__(self) -> str:
        return self.name or f"{date_format(self.date_start)}–{date_format(self.date_end)}"

    class Meta:
        verbose_name = _("Validity range")
        verbose_name_plural = _("Validity ranges")
        constraints = [
            # Heads up: Uniqueness per term implies uniqueness per site
            models.UniqueConstraint(
                fields=["school_term", "date_start", "date_end"], name="lr_unique_dates_per_term"
            ),
        ]
        indexes = [
            models.Index(fields=["date_start", "date_end"], name="lr_validity_date_start_end")
        ]


class Slot(ExtensiblePolymorphicModel):
    """A slot is a time period in which a lesson can take place."""

    WEEKDAY_CHOICES = i18n_day_name_choices_lazy()
    WEEKDAY_CHOICES_SHORT = i18n_day_abbr_choices_lazy()

    validity_range = models.ForeignKey(
        ValidityRange,
        on_delete=models.CASCADE,
        related_name="lr_slots",
        verbose_name=_("Linked validity range"),
    )

    name = models.CharField(verbose_name=_("Name"), max_length=255, blank=True)

    weekday = models.PositiveSmallIntegerField(
        verbose_name=_("Week day"), choices=i18n_day_name_choices_lazy()
    )
    period = models.PositiveSmallIntegerField(
        verbose_name=_("Number of period"), blank=True, null=True
    )

    time_start = models.TimeField(verbose_name=_("Start time"))
    time_end = models.TimeField(verbose_name=_("End time"))

    def __str__(self) -> str:
        if self.name:
            suffix = self.name
        elif self.period:
            suffix = f"{self.period}."
        else:
            suffix = f"{time_format(self.time_start)}–{time_format(self.time_end)}"
        return f"{self.weekday}, {suffix}"

    def get_date(self, week: CalendarWeek) -> date:
        """Get date of lesson in a specific week."""
        return week[self.weekday]

    def get_datetime_start(self, date_ref: Union[CalendarWeek, date]) -> datetime:
        """Get datetime of lesson start in a specific week or on a specific day."""
        if isinstance(date_ref, date):
            day = date_ref
        else:
            day = self.get_date(date_ref)
        return timezone.make_aware(datetime.combine(day, self.time_start))

    def get_datetime_end(self, date_ref: Union[CalendarWeek, int, date]) -> datetime:
        """Get datetime of lesson end in a specific week or on a specific day."""
        if isinstance(date_ref, date):
            day = date_ref
        else:
            day = self.get_date(date_ref)
        return timezone.make_aware(datetime.combine(day, self.time_end))

    class Meta:
        constraints = [
            # Heads up: Uniqueness per validity range implies validity per site
            models.UniqueConstraint(
                fields=["weekday", "period", "validity_range"], name="lr_unique_period_per_range"
            ),
        ]
        ordering = ["weekday", "period"]
        indexes = [models.Index(fields=["time_start", "time_end"])]
        verbose_name = _("Slot")
        verbose_name_plural = _("Slots")


class Lesson(ExtensibleModel):
    """A lesson represents a single teaching event."""

    lesson_event = models.OneToOneField(
        LessonEvent,
        on_delete=models.SET_NULL,
        related_name="lesson",
        verbose_name=_("Linked lesson event"),
        blank=True,
        null=True,
    )

    # A Course is the base of each lesson, it is its planing base

    course = models.ForeignKey(
        Course, on_delete=models.CASCADE, verbose_name=_("Course"), null=True, blank=True
    )

    # Allow lessons to go over multiple slots to ensure
    # that they are later tracked as single event
    slot_start = models.ForeignKey(
        Slot,
        on_delete=models.CASCADE,
        verbose_name=_("Start slot"),
        related_name="+",
    )
    slot_end = models.ForeignKey(
        Slot,
        on_delete=models.CASCADE,
        verbose_name=_("End slot"),
        related_name="+",
    )

    # Recurrence rules allow to define a series of lessons
    # Common examples are weekly or every second week
    recurrence = RecurrenceField(
        verbose_name=_("Recurrence"),
        blank=True,
        null=True,
        help_text=_("Leave empty for a single lesson."),
    )

    # None of the following attributes is required
    # as practice has shown that all possible combinations can occur
    rooms = models.ManyToManyField(
        Room,
        verbose_name=_("Rooms"),
        related_name="lr_lessons",
        blank=True,
    )
    teachers = models.ManyToManyField(
        Person,
        verbose_name=_("Teachers"),
        related_name="lr_lessons_as_teacher",
        blank=True,
    )
    subject = models.ForeignKey(
        Subject,
        on_delete=models.CASCADE,
        verbose_name=_("Subject"),
        related_name="lr_lessons",
        blank=True,
        null=True,
    )

    def clean(self):
        """Ensure that the slots are in the same validity range."""
        if self.slot_start.validity_range != self.slot_end.validity_range:
            raise ValidationError(_("The slots must be in the same validity range."))

    def sync(self):
        """Sync the lesson with its lesson event."""
        week_start = CalendarWeek.from_date(self.slot_start.validity_range.date_start)
        week_end = CalendarWeek.from_date(self.slot_start.validity_range.date_end)
        datetime_start = self.slot_start.get_datetime_start(week_start)
        datetime_end = self.slot_end.get_datetime_end(week_start)
        datetime_end_series = self.slot_end.get_datetime_end(week_end)

        lesson_event, __ = LessonEvent.objects.update_or_create(
            lesson=self,
            defaults={
                "course": self.course,
                "subject": self.subject,
                "datetime_start": datetime_start,
                "datetime_end": datetime_end,
            },
        )
        lesson_event.recurrences = deepcopy(self.recurrence)
        lesson_event.recurrences.exdates += Holiday.get_ex_dates(
            datetime_start, datetime_end_series, self.recurrence
        )
        lesson_event.save()

        self.lesson_event = lesson_event
        self.save()

        lesson_event.groups.set(self.course.groups.all())
        lesson_event.teachers.set(self.teachers.all())
        lesson_event.rooms.set(self.rooms.all())
        return lesson_event

    class Meta:
        # Heads up: Link to slot implies uniqueness per site
        ordering = [
            "slot_start__validity_range__date_start",
            "slot_start__weekday",
            "slot_start__time_start",
            "subject",
        ]
        verbose_name = _("Lesson")
        verbose_name_plural = _("Lessons")


class Break(Slot):
    period_after = models.IntegerField(
        verbose_name=_("Period after"),
    )

    class Meta:
        verbose_name = _("Break")
        verbose_name_plural = _("Breaks")


class Supervision(ExtensibleModel):
    """A supervision is a time period in which a teacher supervises a room."""

    supervision_event = models.OneToOneField(
        SupervisionEvent,
        on_delete=models.SET_NULL,
        related_name="supervision",
        verbose_name=_("Linked supervision event"),
        blank=True,
        null=True,
    )

    rooms = models.ManyToManyField(
        Room,
        verbose_name=_("Rooms"),
        related_name="lr_supervisions",
    )
    teachers = models.ManyToManyField(
        Person,
        verbose_name=_("Teachers"),
        related_name="lr_supervisions",
    )
    slot = models.ForeignKey(
        Break,
        on_delete=models.CASCADE,
        verbose_name=_("Slot"),
        related_name="lr_supervisions",
    )

    # Recurrence rules allow to define a series of supervisions
    # Common examples are weekly or every second week
    recurrence = RecurrenceField(
        verbose_name=_("Recurrence"),
        blank=True,
        null=True,
        help_text=_("Leave empty for a single supervision."),
    )

    class Meta:
        verbose_name = _("Supervision")
        verbose_name_plural = _("Supervisions")

    def sync(self):
        """Sync the supervision with its supervision event."""
        week_start = CalendarWeek.from_date(self.slot.validity_range.date_start)
        week_end = CalendarWeek.from_date(self.slot.validity_range.date_end)
        datetime_start = self.slot.get_datetime_start(week_start)
        datetime_end = self.slot.get_datetime_end(week_start)
        datetime_end_series = self.slot.get_datetime_end(week_end)

        supervision_event, __ = SupervisionEvent.objects.update_or_create(
            supervision=self,
            defaults={
                "datetime_start": datetime_start,
                "datetime_end": datetime_end,
            },
        )

        supervision_event.recurrences = deepcopy(self.recurrence)
        supervision_event.recurrences.exdates += Holiday.get_ex_dates(
            datetime_start, datetime_end_series, self.recurrence
        )
        supervision_event.save()

        self.supervision_event = supervision_event
        self.save()

        supervision_event.teachers.set(self.teachers.all())
        supervision_event.rooms.set(self.rooms.all())
        return supervision_event


class Substitution(ExtensibleModel):
    lesson_event = models.OneToOneField(
        LessonEvent,
        on_delete=models.SET_NULL,
        related_name="substitution",
        verbose_name=_("Linked lesson event"),
        blank=True,
        null=True,
    )

    lesson = models.ForeignKey(Lesson, models.CASCADE, "substitutions", verbose_name=_("Lesson"))

    date = models.DateField(verbose_name=_("Date"))

    rooms = models.ManyToManyField(
        Room,
        verbose_name=_("Rooms"),
        related_name="lr_substitutions",
        blank=True,
    )
    teachers = models.ManyToManyField(
        Person,
        verbose_name=_("Teachers"),
        related_name="lr_substitutions",
        blank=True,
    )
    subject = models.ForeignKey(
        Subject,
        on_delete=models.CASCADE,
        verbose_name=_("Subject"),
        related_name="substitutions",
        blank=True,
        null=True,
    )

    cancelled = models.BooleanField(default=False, verbose_name=_("Cancelled?"))

    comment = models.TextField(verbose_name=_("Comment"), blank=True)

    def clean(self) -> None:
        if self.subject and self.cancelled:
            raise ValidationError(_("Lessons can only be either substituted or cancelled."))

    @property
    def time_range(self) -> (timezone.datetime, timezone.datetime):
        """Get the time range of this substitution."""
        return timezone.datetime.combine(
            self.date, self.lesson_period.period.time_start
        ), timezone.datetime.combine(self.date, self.lesson_period.period.time_end)

    def __str__(self):
        return f"{self.lesson}, {date_format(self.date)}"

    def sync(self):
        """Sync the lesson with its lesson event."""
        week = CalendarWeek.from_date(self.date)
        print(self.lesson.__dict__, self.lesson.lesson_event)
        if not self.lesson.lesson_event:
            return None
        lesson_event, __ = LessonEvent.objects.update_or_create(
            substitution=self,
            defaults={
                "amends": self.lesson.lesson_event,
                "course": self.lesson.course,
                "subject": self.subject,
                "datetime_start": self.lesson.slot_start.get_datetime_start(week),
                "datetime_end": self.lesson.slot_end.get_datetime_end(week),
                "cancelled": self.cancelled,
                "comment": self.comment,
            },
        )
        self.lesson_event = lesson_event
        self.save()
        lesson_event.groups.set(self.lesson.course.groups.all())
        lesson_event.teachers.set(self.teachers.all())
        lesson_event.rooms.set(self.rooms.all())
        print(lesson_event)
        return lesson_event

    class Meta:
        ordering = [
            "date",
            "lesson__slot_start__weekday",
            "lesson__slot_start__period",
        ]
        constraints = [
            models.CheckConstraint(
                check=~Q(cancelled=True, subject__isnull=False),
                name="lr_either_substituted_or_cancelled",
            ),
            # Heads up: Link to period implies uniqueness per site
            models.UniqueConstraint(fields=["lesson", "date"], name="unique_lesson_per_date"),
        ]
        indexes = [
            models.Index(fields=["date"], name="substitution_date"),
            models.Index(fields=["lesson"], name="substitution_lesson"),
        ]
        verbose_name = _("Substitution")
        verbose_name_plural = _("Substitutions")


class SupervisionSubstitution(ExtensibleModel):
    supervision_event = models.OneToOneField(
        SupervisionEvent,
        on_delete=models.SET_NULL,
        related_name="supervision_substitution",
        verbose_name=_("Linked supervision event"),
        blank=True,
        null=True,
    )

    supervision = models.ForeignKey(
        Supervision, models.CASCADE, "substitutions", verbose_name=_("Supervision")
    )

    date = models.DateField(verbose_name=_("Date"))

    teachers = models.ManyToManyField(
        Person,
        verbose_name=_("Teachers"),
        related_name="lr_supervision_substitutions",
        blank=True,
    )

    cancelled = models.BooleanField(default=False, verbose_name=_("Cancelled?"))

    comment = models.TextField(verbose_name=_("Comment"), blank=True)

    @property
    def time_range(self) -> (timezone.datetime, timezone.datetime):
        """Get the time range of this supervision."""
        return timezone.datetime.combine(
            self.date, self.supervision.slot.period.time_start
        ), timezone.datetime.combine(self.date, self.supervision.slot.period.time_end)

    def __str__(self):
        return f"{self.supervision}, {date_format(self.date)}"

    def sync(self):
        """Sync the supervision with its supervision event."""
        week = CalendarWeek.from_date(self.date)
        if not self.supervision.supervision_event:
            return None
        supervision_event, __ = SupervisionEvent.objects.update_or_create(
            supervision_substitution=self,
            defaults={
                "amends": self.supervision.supervision_event,
                "datetime_start": self.supervision.slot.get_datetime_start(week),
                "datetime_end": self.supervision.slot.get_datetime_end(week),
                "cancelled": self.cancelled,
                "comment": self.comment,
            },
        )
        self.supervision_event = supervision_event
        self.save()
        supervision_event.teachers.set(self.teachers.all())
        return supervision_event

    class Meta:
        ordering = [
            "date",
            "supervision__slot__weekday",
            "supervision__slot__period",
        ]
        indexes = [
            models.Index(fields=["date"], name="super_substitution_date"),
            models.Index(fields=["supervision"], name="super_substitution_supervision"),
        ]
        verbose_name = _("Supervision Substitution")
        verbose_name_plural = _("Supervision Substitutions")


class TimeboundCourseConfig(ExtensibleModel):
    """A timebound course config is the specific configuration for a combination of a course and a validity range."""

    course = models.ForeignKey(
        Course,
        on_delete=models.CASCADE,
        verbose_name=_("Course"),
        related_name="lr_timebound_course_configs",
    )
    validity_range = models.ForeignKey(
        ValidityRange,
        on_delete=models.CASCADE,
        verbose_name=_("Linked validity range"),
        related_name="lr_timebound_course_configs",
    )

    scheduled_slot_count = models.PositiveSmallIntegerField(
        verbose_name=_("Number of slots this course is scheduled to fill per week"),
        blank=True,
        null=True,
    )
    teachers = models.ManyToManyField(
        Person,
        verbose_name=_("Teachers"),
        related_name="lr_timebound_course_configs",
    )

    class Meta:
        constraints = [
            # Heads up: Uniqueness per validity range implies validity per site
            models.UniqueConstraint(
                fields=["course", "validity_range"], name="lr_unique_course_config_per_range"
            ),
        ]
        verbose_name = _("Timebound course config")
        verbose_name_plural = _("Timebound course configs")
