from django.db.models import Prefetch

import graphene
from guardian.shortcuts import get_objects_for_user

from aleksis.apps.cursus.models import Course, Subject
from aleksis.apps.cursus.schema import Query as CursusSchemaQuery
from aleksis.core.models import Group, Person
from aleksis.core.schema.base import FilterOrderList
from aleksis.core.schema.group import GroupType
from aleksis.core.schema.person import PERSON_TYPE_FIELDS, PersonType

from ..models import Slot, ValidityRange
from .break_type import (
    BreakSlotBatchCreateMutation,
    BreakSlotBatchDeleteMutation,
    BreakSlotBatchPatchMutation,
    BreakSlotCreateMutation,
    BreakSlotDeleteMutation,
    BreakSlotType,
)
from .slot import (
    CarryOverSlotsMutation,
    SlotBatchCreateMutation,
    SlotBatchDeleteMutation,
    SlotBatchPatchMutation,
    SlotCreateMutation,
    SlotDeleteMutation,
    SlotType,
)
from .timebound_course_config import (
    LesroosterExtendedSubjectType,
    TimeboundCourseConfigBatchCreateMutation,
    TimeboundCourseConfigBatchPatchMutation,
    TimeboundCourseConfigCreateMutation,
    TimeboundCourseConfigDeleteMutation,
    TimeboundCourseConfigType,
)
from .validity_range import (
    ValidityRangeBatchDeleteMutation,
    ValidityRangeBatchPatchMutation,
    ValidityRangeCreateMutation,
    ValidityRangeDeleteMutation,
    ValidityRangeType,
)


class LesroosterExtendedPersonType(PersonType):
    class Meta:
        model = Person
        fields = PERSON_TYPE_FIELDS + ["lesson_quota"]


class Query(graphene.ObjectType):
    break_slots = FilterOrderList(BreakSlotType)
    slots = FilterOrderList(SlotType)
    timebound_course_configs = FilterOrderList(TimeboundCourseConfigType)
    validity_ranges = FilterOrderList(ValidityRangeType)

    current_validity_range = graphene.Field(ValidityRangeType)

    lesrooster_extended_subjects = FilterOrderList(
        LesroosterExtendedSubjectType, courses=graphene.List(graphene.ID)
    )

    classes_by_grade = graphene.List(GroupType, grade=graphene.ID())

    lesrooster_extended_teachers = FilterOrderList(LesroosterExtendedPersonType)

    @staticmethod
    def resolve_slots(root, info):
        # Note: This does also return `Break` objects (but with type set to Slot). This is intended
        return Slot.objects.non_polymorphic()

    @staticmethod
    def resolve_time_course_configs(root, info):
        return

    @staticmethod
    def resolve_lesrooster_extended_subjects(root, info, courses):
        return get_objects_for_user(
            info.context.user, "cursus.view_subject", Subject.objects.all()
        ).prefetch_related(
            Prefetch(
                "courses",
                queryset=get_objects_for_user(
                    info.context.user, "cursus.view_course", Course.objects.all()
                ).filter(groups__in=courses),
            )
        )

    @staticmethod
    def resolve_current_validity_range(root, info):
        return ValidityRange.current

    @staticmethod
    def resolve_classes_by_grade(root, info, grade):
        grade_group = Group.objects.prefetch_related("child_groups").get(id=grade)
        return get_objects_for_user(
            info.context.user,
            "core.view_group",
            grade_group.child_groups.filter(group_type__name="School class"),
        )

    @staticmethod
    def resolve_lesrooster_extended_teachers(root, info):
        return CursusSchemaQuery.resolve_teachers(root, info)


class Mutation(graphene.ObjectType):
    create_break_slot = BreakSlotCreateMutation.Field()
    create_break_slots = BreakSlotBatchCreateMutation.Field()
    delete_break_slot = BreakSlotDeleteMutation.Field()
    delete_break_slots = BreakSlotBatchDeleteMutation.Field()
    update_break_slots = BreakSlotBatchPatchMutation.Field()

    create_slot = SlotCreateMutation.Field()
    create_slots = SlotBatchCreateMutation.Field()
    delete_slot = SlotDeleteMutation.Field()
    delete_slots = SlotBatchDeleteMutation.Field()
    update_slots = SlotBatchPatchMutation.Field()

    batch_create_timebound_course_config = TimeboundCourseConfigBatchCreateMutation.Field()
    create_timebound_course_config = TimeboundCourseConfigCreateMutation.Field()
    delete_timebound_course_config = TimeboundCourseConfigDeleteMutation.Field()
    update_timebound_course_configs = TimeboundCourseConfigBatchPatchMutation.Field()
    carry_over_slots = CarryOverSlotsMutation.Field()

    create_validity_range = ValidityRangeCreateMutation.Field()
    delete_validity_range = ValidityRangeDeleteMutation.Field()
    delete_validity_ranges = ValidityRangeBatchDeleteMutation.Field()
    update_validity_ranges = ValidityRangeBatchPatchMutation.Field()
