from django.db.models import Q

import graphene
from graphene_django.types import DjangoObjectType
from graphene_django_cud.mutations import (
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoBatchPatchMutation,
    DjangoCreateMutation,
)

from aleksis.core.schema.base import (
    DeleteMutation,
    DjangoFilterMixin,
    PermissionBatchDeleteMixin,
    PermissionBatchPatchMixin,
    PermissionsTypeMixin,
)

from ..models import Break, Slot, ValidityRange

slot_filters = {
    "id": ["exact"],
    "name": ["exact", "icontains"],
    "weekday": ["exact", "in"],
    "period": ["exact", "lt", "lte", "gt", "gte"],
    "time_start": ["exact", "lt", "lte", "gt", "gte"],
    "time_end": ["exact", "lt", "lte", "gt", "gte"],
    "validity_range": ["exact"],
}


class SlotType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    model = graphene.String(default_value="Default")
    period_after = graphene.Int()

    class Meta:
        model = Slot
        fields = ("id", "validity_range", "name", "weekday", "period", "time_start", "time_end")
        filter_fields = slot_filters

    @classmethod
    def get_queryset(cls, queryset, info):
        return queryset  # FIXME filter this queryset based on permissions

    @staticmethod
    def resolve_model(root, info):
        return root.get_real_instance_class().__name__

    @staticmethod
    def resolve_period_after(root, info):
        return getattr(root.get_real_instance(), "period_after", None)


class SlotCreateMutation(DjangoCreateMutation):
    class Meta:
        model = Slot
        field_types = {"weekday": graphene.Int()}
        permissions = ("",)  # FIXME


class SlotDeleteMutation(DeleteMutation):
    klass = Slot
    permission_required = ""  # FIXME


class SlotBatchCreateMutation(PermissionBatchPatchMixin, DjangoBatchCreateMutation):
    class Meta:
        model = Slot
        field_types = {"weekday": graphene.Int()}
        permissions = ("",)  # FIXME


class SlotBatchDeleteMutation(PermissionBatchDeleteMixin, DjangoBatchDeleteMutation):
    class Meta:
        model = Slot
        permissions = ("lesrooster.delete_slot",)


class SlotBatchPatchMutation(PermissionBatchPatchMixin, DjangoBatchPatchMutation):
    class Meta:
        model = Slot
        field_types = {"weekday": graphene.Int()}
        permissions = ("",)  # FIXME


class CarryOverSlotsMutation(graphene.Mutation):
    class Arguments:
        validity_range = graphene.ID()
        from_day = graphene.Int()
        to_day = graphene.Int()

        only = graphene.List(graphene.ID, required=False)

    deleted = graphene.List(graphene.ID)
    result = graphene.List(SlotType)

    @classmethod
    def mutate(cls, root, info, validity_range, from_day, to_day, only=None):
        if only is None:
            only = []

        # TODO: Check permissions

        validity_range = ValidityRange.objects.get(id=validity_range)

        slots_on_day = Slot.objects.filter(weekday=from_day, validity_range=validity_range)

        if only and len(only) > 0:
            slots_on_day = slots_on_day.filter(id__in=only)

        periods = []
        periods_after = []

        for slot in slots_on_day:
            slot: Break | Slot
            klass = slot.get_real_instance_class()

            defaults = {"name": slot.name, "time_start": slot.time_start, "time_end": slot.time_end}

            if hasattr(slot, "period_after"):
                defaults["period_after"] = slot.period_after
                periods_after.append(slot.period_after)

            if slot.period is not None:
                periods.append(slot.period)

            klass.objects.update_or_create(
                weekday=to_day, validity_range=validity_range, period=slot.period, defaults=defaults
            )

        if not only or not len(only):
            objects_to_delete = Slot.objects.filter(
                weekday=to_day, validity_range=validity_range
            ).exclude(Q(Slot___period__in=periods) | Q(Break___period_after__in=periods_after))
            objects_to_delete.delete()

            deleted = objects_to_delete.values_list("id", flat=True)

        else:
            deleted = []

        return CarryOverSlotsMutation(
            deleted=deleted,
            result=Slot.objects.filter(
                weekday=to_day, validity_range=validity_range
            ).non_polymorphic(),
        )
